"""
Objects for YAML serializer/deserializer
"""
import os
import json
import logging
from pathlib import PurePosixPath

import yaml
from unidecode import unidecode

from django.core.files import File
from django.db import transaction
from wagtail.wagtailcore.models import Page as WagtailPage
from wagtail.wagtailcore.fields import StreamField
from wagtail.wagtailimages.models import Image as WagtailImage

LOGGER = logging.getLogger(__name__)


class JSONSerializable(object):
    """Interface to objects which are serializable into JSON."""

    def __to_json__(self):
        """Serialize to JSON."""
        raise NotImplementedError()


class FieldStorable(object):
    """Interface to objects which can be stored in Django model fields."""

    def __to_value__(self):
        """Value to store in a Django model."""
        raise NotImplementedError()


class JSONEncoder(json.JSONEncoder):
    """
    Extension of the JSON encoder that knows how to encode the YAML objects
    """

    def default(self, obj):  # pylint:disable=method-hidden
        if isinstance(obj, JSONSerializable):
            return obj.__to_json__()

        return super().default(obj)


class GetForeignObject(FieldStorable, yaml.YAMLObject):
    """
    Get a foreign key reference for the provided parameters
    """

    yaml_loader = yaml.SafeLoader

    @property
    def model(self):
        """Model for this reference."""
        raise NotImplementedError()

    @property
    def lookup_keys(self):
        """
        Lookup keys.

        Default implementation is all model fields that are defined in
        the Yaml block.
        """

        return (field.name for field in self.model._meta.get_fields())

    def lookup(self):
        """
        Generate the lookup
        """

        return {
            field: getattr(self, field)
            for field in self.lookup_keys
            if hasattr(self, field)
        }

    def get_object(self):
        """
        Get the object from the database.
        """
        return self.model.objects.get(**self.lookup())

    def __to_value__(self):
        obj = self.get_object()

        # update the object with any remaining keys
        for field in self.model._meta.get_fields():
            if hasattr(self, field.name):
                value = getattr(self, field.name)

                if isinstance(field, StreamField):
                    value = json.dumps(value, cls=JSONEncoder)

                elif isinstance(value, FieldStorable):
                    value = value.__to_value__()  # pylint:disable=no-member

                setattr(obj, field.name, value)

        return obj


# pylint:disable=abstract-method
class GetOrCreateForeignObject(GetForeignObject):
    """
    Get or create a foreign key reference for the provided parameters
    """

    def get_object(self):
        obj, _ = self.model.objects.get_or_create(**self.lookup())
        return obj
# pylint:enable=abstract-method


class Page(JSONSerializable, yaml.YAMLObject):
    """
    A reference to a page
    """

    yaml_tag = '!page'
    yaml_loader = yaml.SafeLoader

    def __to_json__(self):
        # pylint:disable=no-member
        url = PurePosixPath(self.url)

        if not url.is_absolute():
            raise ValueError("URL must be absolute")

        page = WagtailPage.objects.only('id').get(url_path=str(url) + '/')

        return page.id


class Image(FieldStorable, JSONSerializable, yaml.YAMLObject):
    """
    A reference to an image
    """

    yaml_tag = '!image'
    yaml_loader = yaml.SafeLoader

    file = None  # expected parameter

    @property
    def db_filename(self):
        """
        Filename that will be stored into the DB.

        This code is taken from Wagtail. We can't call into the Wagtail code
        because it will create unique filenames.
        """

        folder_name = 'original_images'
        filename = ('images/%s' % self.file).replace('/', '')
        filename = "".join((i if ord(i) < 128 else '_')
                           for i in unidecode(filename))

        while len(os.path.join(folder_name, filename)) >= 95:
            prefix, dot, extension = filename.rpartition('.')
            filename = prefix[:-1] + dot + extension

        return os.path.join(folder_name, filename)

    @transaction.atomic
    def __to_value__(self):
        try:
            obj = WagtailImage.objects.get(file=self.db_filename)

        except WagtailImage.DoesNotExist:
            print("Creating file %s..." % self.db_filename)
            with open('images/%s' % self.file, 'rb') as file_:
                file_ = File(file_)
                obj = WagtailImage(file=file_)
                # Need to save while the file is open
                obj.save()  # pylint:disable=no-member

        # copy the fields
        for field in obj._meta.get_fields():
            if field.name == 'file':
                continue
            elif hasattr(self, field.name):
                setattr(obj, field.name, getattr(self, field.name))

        obj.save()

        return obj

    def __to_json__(self):
        return self.__to_value__().id



"""
Import pages into Wagtail
"""
import os
import json
from pathlib import PurePosixPath, Path

import yaml

from django.db import transaction
from django.core.exceptions import FieldDoesNotExist
from django.core.management.base import BaseCommand, CommandError
from django.contrib.contenttypes.models import ContentType

from wagtail.wagtailcore.models import Page
from wagtail.wagtailcore.fields import StreamField

from ... import serializer


class Command(BaseCommand):
    """
    Import pages into Wagtail.
    """

    def add_arguments(self, parser):
        parser.add_argument('file', nargs='+', type=str)

    @transaction.atomic
    def handle(self, *args, **options):
        for filename in options['file']:
            with open(filename) as file_:
                docs = yaml.safe_load_all(file_)
                self.stdout.write("Reading %s" % filename)

                cwd = Path.cwd()
                try:
                    os.chdir(str(Path(filename).parent))
                    self.import_documents(docs)
                finally:
                    os.chdir(str(cwd))

    @transaction.atomic
    def import_documents(self, docs):
        """Import a Yaml file of documents."""

        for doc in docs:
            try:
                if isinstance(doc, serializer.GetForeignObject):
                    self.import_snippet(doc)
                else:
                    self.import_page(doc)
            except CommandError as exc:
                self.stderr.write("Error importing page: %s" % exc)

    @transaction.atomic
    def import_snippet(self, data):
        """Import a snippet (which is a GetForeignObject)."""

        obj = data.__to_value__()
        print("Importing snippet %s" % obj)
        obj.save()

    @transaction.atomic
    def import_page(self, data):
        """Import a single wagtail page."""
        model = self.get_page_model_class(data)
        self.find_page(model, data)

    def get_page_model_class(self, data):
        """
        Get the page model class from the `type' parameter.
        """

        try:
            type_ = data.pop('type')
        except KeyError:
            raise CommandError("Need `type' for page")

        try:
            app_label, model = type_.split('.')
            return ContentType.objects.get(app_label=app_label,
                                           model=model)\
                .model_class()
        except (ValueError, AttributeError):
            raise CommandError("`type' is of form `app.model'")
        except ContentType.DoesNotExist:
            raise CommandError("Unknown page type `%s'" % type_)

    def find_page(self, model, data):
        """
        Find a page by its URL and import its data.

        Data importing has to be done here because often the page can't
        be saved until the data is imported (i.e. null fields)
        """
        try:
            url = PurePosixPath(data.pop('url'))
            if not url.is_absolute():
                raise CommandError("Path %s must be absolute" % url)

        except KeyError:
            raise CommandError("Need `url' for page")

        try:
            page = model.objects.get(url_path=str(url) + '/')
            self.import_data(page, data)
            page.save()
            self.stdout.write("Updating existing page %s" % url)
        except model.DoesNotExist:
            try:
                # pylint:disable=no-member
                parent = Page.objects.get(url_path=str(url.parent) + '/')
            except Page.DoesNotExist:
                raise CommandError("Parent of %s doesn't exist" % url)

            page = model(slug=url.name)
            self.import_data(page, data)
            parent.add_child(instance=page)
            self.stdout.write("Creating new page %s" % url)

        return page

    def import_data(self, page, data):
        """Import the data onto a page."""

        for key, value in data.items():
            try:
                field = page._meta.get_field(key)

                if isinstance(field, StreamField):
                    value = json.dumps(value, cls=serializer.JSONEncoder)
                else:
                    # Assume we know how to serialise it
                    pass

            except FieldDoesNotExist:
                # This might be a property, just try and set it anyway
                pass

            if isinstance(value, serializer.FieldStorable):
                value = value.__to_value__()  # pylint:disable=no-member

            setattr(page, key, value)



"""
setuptools configuration for wagtailimporter.
"""

from setuptools import setup, find_packages

with open('README.rst') as readme, \
     open('requirements.txt') as requirements:
    setup(
        name="wagtailimporter",
        version="0.0.10",
        description="Wagtail module to load pages from Yaml",
        long_description=readme.read(),
        url='https://github.com/squareweave/wagtailimporter',
        author='Squareweave',
        author_email='team@squareweave.com.au',
        license='BSD',
        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Developers',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
        ],
        packages=find_packages(),
        include_package_data=True,
        install_requires=requirements.readlines(),
    )
